"""Placement simulation package, contains the placement simulators."""
from .pymunk import Chipmunk

try:
    from .pybox2d import Box2D
except ImportError:
    Box2D = None
from .base import PlacementSimulation

__all__ = ['Chipmunk', 'Box2D', 'PlacementSimulation']
