""" CellSium - _Cell_ _Si_mulator for _micro_fluidic _m_icrocolonies """

__project__ = 'CellSium'

__version__ = '1.0.0'

__author__ = 'Christian C. Sachs'

__copyright__ = '2021, Christian C. Sachs, Forschungszentrum Jülich GmbH'

__citation__ = '''
Sachs CC, Ruzaeva K, Seiffarth J, Wiechert W, Berkels B, Nöh K (2022)
CellSium: versatile cell simulator for microcolony ground truth generation
Bioinformatics Advances, Volume 2, Issue 1, 2022, vbac053, doi: 10.1093/bioadv/vbac053
'''

__banner__ = ''
